class Auth:
    def __init__(
        self,
        token=None,
    ):
        self.token = token

    async def check_token(self):
        raise NotImplementedError

    async def login(self, email: str):
        raise NotImplementedError



from datetime import datetime
import logging
from typing import Any, Dict

from . import abc
from ..state import ConnectionState
from .letter import AsyncLetterIterator

log = logging.getLogger(__name__)


class BaseUser(abc.User):
    __slots__ = (
        "_state",
        "dob",
        "age",
        "allowaudio",
        "allowphotos",
        "audiorequest",
        "avatar",
        "by_id",
        "created_at",
        "customdesc",
        "deactivated",
        "dob_privacy",
        "emoji_status",
        "fav",
        "gender",
        "id",
        "identity",
        "joined",
        "joined_at",
        "joined_audio",
        "joined_photos",
        "latest_comment",
        "latest_sent_by",
        "location_code",
        "name",
        "openletter",
        "photorequest",
        "plus",
        "show_last_login",
        "status",
        "total",
        "unread",
        "updated_at",
        "user_audio",
        "user_id",
        "user_photos",
        "user_status",
    )

    def __init__(self, state: ConnectionState, *, data: Dict[str, Any]) -> None:
        """
        Initialize a BaseUser instance.

        :param state: The connection state.
        :param data: The data to initialize the user with.
        """
        self._state: ConnectionState = state
        self._update(data)

    def __str__(self) -> str:
        """
        Return the string representation of the user.

        :return: The name of the user.
        """
        return self.name

    def _update(self, data: Dict[str, Any]) -> None:
        """
        Update the user attributes with the provided data.

        :param data: The data to update the user with.
        """
        for attr in BaseUser.__slots__:
            if attr.startswith("_"):
                continue
            if attr in [
                "created_at",
                "latest_comment",
                "updated_at",
                "joined_at",
            ] and data.get(attr):
                date_object = datetime.strptime(data[attr], "%Y-%m-%d %H:%M:%S")
                setattr(self, attr, date_object)
                continue
            elif attr == "dob" and data.get("dob"):
                date_object = datetime.strptime(data["dob"], "%Y-%m-%d")
                setattr(self, attr, date_object)
                continue
            setattr(self, attr, data.get(attr, None))

    def __repr__(self) -> str:
        """
        Return the official string representation of the user.

        :return: The official string representation of the user.
        """
        return "<User name={0.name!r} id={0.id!r}>".format(self)


class User(BaseUser):
    def __init__(self, state: ConnectionState, *, data: Dict[str, Any]) -> None:
        """
        Initialize a User instance.

        :param state: The connection state.
        :param data: The data to initialize the user with.
        """
        super().__init__(state, data=data)

    def letters(self) -> AsyncLetterIterator:
        """
        Return an iterator for the user's letters.

        :return: An AsyncLetterIterator instance.
        """
        return AsyncLetterIterator(self._state, self.id)



import gc


class ConnectionState:
    def __init__(self, *, dispatch, handlers, http, loop, **options):
        self.loop = loop
        self.http = http
        self.dispatch = dispatch
        self.handlers = handlers
        self.heartbeat_timeout = options.get("heartbeat_timeout", 60.0)

    def clear(self):
        self.client_user = None
        self._users = {}
        self._messages = []
        gc.collect()

    def call_handlers(self, key, *args, **kwargs):
        try:
            func = self.handlers[key]
        except KeyError:
            pass
        else:
            func(*args, **kwargs)

    def store_user(self, user):
        try:
            return self._users[user.id]
        except KeyError:
            self._users[user.id] = user
            return user



__title__ = "tinder"
__author__ = "tiagovla"
__version__ = "0.1.2"
__license__ = "GPL-3.0-or-later"

import logging

from .client import Client

__all__ = ["Client"]

from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

del logging



class SlowlyException(Exception):
    pass


class ClientException(SlowlyException):
    pass


class HTTPException(SlowlyException):
    def __init__(self, response, message):
        self.response = response
        self.status = response.status
        self.text = message

        if isinstance(message, dict):
            self.text = message.get("error")

        fmt = "Status: {0.status} Error: {1}"
        super().__init__(fmt.format(self.response, message))


class Forbidden(HTTPException):
    pass


class NotFound(HTTPException):
    pass


class InvalidData(ClientException):
    pass


class LoginFailure(ClientException):
    pass



import asyncio
import logging
import signal
from typing import Any, Callable, Dict, List, Optional, Tuple
from .models import User
from .http import HTTPClient
from .state import ConnectionState

log = logging.getLogger(__name__)


class _ClientEventTask(asyncio.Task):
    def __init__(
        self,
        original_coro: Callable,
        event_name: str,
        coro: Callable,
        *,
        loop: asyncio.AbstractEventLoop
    ) -> None:
        super().__init__(coro, loop=loop)
        self.__event_name = event_name
        self.__original_coro = original_coro

    def __repr__(self) -> str:
        info = [
            ("state", self._state.lower()),
            ("event", self.__event_name),
            ("coro", repr(self.__original_coro)),
        ]
        if self._exception is not None:
            info.append(("exception", repr(self._exception)))
        return "<ClientEventTask {}>".format(" ".join("%s=%s" % t for t in info))


class Client:
    def __init__(
        self, *, loop: Optional[asyncio.AbstractEventLoop] = None, **options: Any
    ) -> None:
        self.loop = loop or asyncio.get_event_loop()
        self.connector = options.pop("connector", None)
        self.proxy = options.pop("proxy", None)
        self.proxy_auth = options.pop("proxy_auth", None)
        self._listeners: Dict[str, List[Tuple[asyncio.Future, Callable]]] = {}
        self.http = HTTPClient(
            self.connector, proxy=self.proxy, proxy_auth=self.proxy_auth, loop=self.loop
        )
        self._ready = asyncio.Event()
        self._handlers = {"ready": self._handle_ready}
        self._connection = ConnectionState(
            dispatch=self.dispatch,
            handlers=self._handlers,
            http=self.http,
            loop=self.loop,
            **options,
        )

    def is_ready(self) -> bool:
        """Check if the client is ready."""
        return self._ready.is_set()

    async def _run_event(
        self, coro: Callable, event_name: str, *args: Any, **kwargs: Any
    ) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            log.error("Exception in event '%s': %s", event_name, e)
            try:
                await self.on_error(event_name, *args, **kwargs)
            except asyncio.CancelledError:
                pass

    async def on_error(self, event_method: str, *args: Any, **kwargs: Any) -> None:
        """Handle errors during event execution."""
        log.error("Ignoring exception in %s", event_method, exc_info=True)

    def _schedule_event(
        self, coro: Callable, event_name: str, *args: Any, **kwargs: Any
    ) -> _ClientEventTask:
        wrapped = self._run_event(coro, event_name, *args, **kwargs)
        return _ClientEventTask(
            original_coro=coro, event_name=event_name, coro=wrapped, loop=self.loop
        )

    async def wait_until_ready(self) -> None:
        """Wait until the client is ready."""
        await self._ready.wait()

    def wait_for(
        self,
        event: str,
        *,
        check: Optional[Callable] = None,
        timeout: Optional[float] = None
    ) -> asyncio.Future:
        """Wait for a specific event to occur."""
        future = self.loop.create_future()
        check = check if check else lambda *args: True
        ev = event.lower()
        try:
            listeners = self._listeners[ev]
        except KeyError:
            listeners = []
            self._listeners[ev] = listeners

        listeners.append((future, check))
        return asyncio.wait_for(future, timeout)

    def dispatch(self, event: str, *args: Any, **kwargs: Any) -> None:
        """Dispatch an event to the appropriate handlers."""
        log.debug(
            "Dispatching event: %s with args: %s and kwargs: %s", event, args, kwargs
        )
        method = "on_" + event

        listeners = self._listeners.get(event)
        if listeners:
            removed = []
            for i, (future, condition) in enumerate(listeners):
                if future.cancelled():
                    removed.append(i)
                    continue

                try:
                    result = condition(*args)
                except Exception as exec:
                    log.error(
                        "Error in event listener condition for event '%s': %s",
                        event,
                        exec,
                    )
                    future.set_exception(exec)
                    removed.append(i)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(i)

                if len(removed) == len(listeners):
                    self._listeners.pop(event)
                else:
                    for idx in reversed(removed):
                        del listeners[idx]

        try:
            coro = getattr(self, method)
        except AttributeError:
            log.warning("No handler found for event: %s", event)
        else:
            self._schedule_event(coro, method, *args, **kwargs)

    async def close(self) -> None:
        """Close the client."""
        log.debug("Closing client")
        await self.http.close()

    async def login(self, token: str) -> None:
        """Login to the client using a token."""
        log.debug("Logging in with token: %s", token)
        await self.http.login(token.strip())

    async def start(self, *args: Any, **kwargs: Any) -> None:
        """Start the client."""
        log.debug("Starting client with args: %s and kwargs: %s", args, kwargs)
        await self.login(*args)

    async def main(self) -> None:
        """Main function to be overridden by subclasses."""
        pass

    def _handle_ready(self) -> None:
        """Handle the ready event."""
        log.debug("Ready event triggered")
        self._ready.set()

    def run(self, *args: Any, **kwargs: Any) -> Optional[Any]:
        """Run the client."""
        loop = self.loop
        try:
            loop.add_signal_handler(signal.SIGINT, lambda: loop.stop())
            loop.add_signal_handler(signal.SIGTERM, lambda: loop.stop())
        except NotImplementedError:
            pass

        async def runner() -> None:
            try:
                await self.start(*args, **kwargs)
                await self.main()
            finally:
                await self.close()

        def stop_loop_on_completion(f: asyncio.Future) -> None:
            loop.stop()

        future = asyncio.ensure_future(runner(), loop=loop)
        future.add_done_callback(stop_loop_on_completion)
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            log.info("Received terminate signal")
        finally:
            future.remove_done_callback(stop_loop_on_completion)

        if not future.cancelled():
            try:
                return future.result()
            except KeyboardInterrupt:
                return None

    def event(self, coro: Callable) -> Callable:
        """Register an event coroutine."""
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("event registered must be a coroutine function")

        setattr(self, coro.__name__, coro)
        log.debug("Event %s has been successfully registered", coro.__name__)
        return coro

    async def fetch_client_profile(self) -> None:
        """Fetch the client profile."""
        log.debug("Fetching client profile")
        await self.http.fetch_client_profile()

    async def fetch_friends(self) -> List[User]:
        """Fetch the list of friends."""
        data = await self.http.fetch_friends()
        log.debug("Fetched friends data: %s", data)
        return [User(self._connection, data=friend) for friend in data["friends"]]

    async def fetch_passcode(self, email: str) -> None:
        """Fetch the passcode for the given email."""
        log.debug("Fetching passcode for email: %s", email)
        await self.http.fetch_auth_passcode(email)

    async def fetch_token(self, email: str, passcode: str) -> str:
        """Fetch the token for the given email and passcode."""
        log.debug("Fetching token for email: %s with passcode: %s", email, passcode)
        response = await self.http.fetch_auth_token(email, passcode)
        log.debug("Fetched token: %s", response["token"])
        return response["token"]



import asyncio
import uuid
import logging
from typing import Optional, Any, Dict, Coroutine
from urllib.parse import quote as _uriquote

import aiohttp
from .errors import Forbidden, HTTPException, NotFound

log: logging.Logger = logging.getLogger(__name__)


async def json_or_text(response: aiohttp.ClientResponse) -> Dict[str, Any] | str:
    """
    Parse the response as JSON if possible, otherwise return as text.

    :param response: The response object to parse.
    :type response: aiohttp.ClientResponse
    :return: The parsed JSON data or text.
    :rtype: dict[str, Any] or str
    """
    try:
        if "application/json" in response.headers["content-type"]:
            return await response.json()
    except KeyError:
        pass
    return await response.text(encoding="utf-8")


class Route:
    BASE = "https://api.getslowly.com/"

    def __init__(self, method: str, path: str, **params: Any) -> None:
        """
        Initialize a Route instance.

        :param method: The HTTP method (e.g., 'GET', 'POST').
        :type method: str
        :param path: The API endpoint path.
        :type path: str
        :param params: Additional parameters for the URL.
        :type params: Any
        """
        self.path = path
        self.method = method
        url: str = self.BASE + self.path
        if params:
            self.url = url.format(
                **{
                    k: _uriquote(v) if isinstance(v, str) else v
                    for k, v in params.items()
                }
            )
        else:
            self.url = url


class HTTPClient:
    def __init__(
        self,
        connector: Optional[aiohttp.BaseConnector] = None,
        *,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        """
        Initialize an HTTPClient instance.

        :param connector: The aiohttp connector to use, by default None.
        :type connector: Optional[aiohttp.BaseConnector], optional
        :param proxy: The proxy URL, by default None.
        :type proxy: Optional[str], optional
        :param proxy_auth: The proxy authentication, by default None.
        :type proxy_auth: Optional[aiohttp.BasicAuth], optional
        :param loop: The event loop to use, by default None.
        :type loop: Optional[asyncio.AbstractEventLoop], optional
        """
        self.loop: asyncio.AbstractEventLoop = loop or asyncio.get_event_loop()
        self.connector: Optional[aiohttp.BaseConnector] = connector
        self.__session: aiohttp.ClientSession
        self.token: Optional[str] = None
        self.proxy: Optional[str] = proxy
        self.proxy_auth: Optional[aiohttp.BasicAuth] = proxy_auth
        self.__global_over: asyncio.Event = asyncio.Event()
        self.__global_over.set()
        self.user_agent: str = " ".join(
            [
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64)",
                "AppleWebKit/537.36 (KHTML, like Gecko)",
                "Chrome/132.0.0.0 Safari/537.36",
            ]
        )
        self.device = {
            "uuid": str(uuid.uuid5(uuid.NAMESPACE_DNS, str(uuid.getnode()))),
            "os": "Linux x86_64",
            "browser": "Chrome 132",
            "locale": "en",
            "trusted": "true",
            "version": "4.0.x",
        }

    async def request(self, route: Route, **kwargs: Any) -> dict[str, Any] | str:
        """
        Make an HTTP request.

        :param route: The route object containing method and URL.
        :type route: Route
        :param kwargs: Additional arguments for the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: dict[str, Any] or str
        :raises Forbidden: If the response status is 403.
        :raises NotFound: If the response status is 404.
        :raises HTTPException: For other HTTP errors.
        :raises RuntimeError: If the code is unreachable.
        """
        method = route.method
        url = route.url
        headers: Optional[dict[str, str]] = kwargs.get("headers")
        if headers is None:
            headers = {
                "accept": "application/json",
                "content-type": "application/json",
                "origin": "https://web.slowly.app",
                "user-agent": self.user_agent,
            }
        if self.token:
            headers["authorization"] = f"Bearer {self.token}"
        if "json" in kwargs:
            headers["Content-Type"] = "application/json"
        kwargs["headers"] = headers
        if self.proxy:
            kwargs["proxy"] = self.proxy
        elif self.proxy_auth:
            kwargs["proxy_auth"] = self.proxy_auth
        if not self.__global_over.is_set():
            await self.__global_over.wait()
        for tries in range(3):
            async with self.__session.request(method, url, **kwargs) as r:
                data: dict[str, Any] | str = await json_or_text(r)
                if 300 > r.status >= 200:
                    return data
                elif r.status in {500, 502}:
                    await asyncio.sleep(1 + tries * 2)
                    continue
                elif r.status == 403:
                    raise Forbidden(r, data)
                elif r.status == 404:
                    raise NotFound(r, data)
                else:
                    raise HTTPException(r, data)
        raise RuntimeError("Unreachable code in HTTP handling")

    async def login(self, token: str) -> None:
        """
        Login with the provided token.

        :param token: The authentication token.
        :type token: str
        """
        log.debug("Logging in with token: %s", token)
        self.__session = aiohttp.ClientSession(connector=self.connector)
        self.token = token

    async def close(self) -> None:
        """
        Close the HTTP session.
        """
        log.debug("Closing the HTTP session")
        if self.__session:
            await self.__session.close()

    async def recreate(self) -> None:
        """
        Recreate the HTTP session if it is closed.
        """
        log.debug("Recreating the HTTP session")
        if self.__session and self.__session.closed:
            self.__session = aiohttp.ClientSession(connector=self.connector)

    def fetch_client_profile(self) -> Coroutine[Any, Any, dict[str, Any] | str]:
        """
        Get the client's profile.

        :return: The coroutine for the request.
        :rtype: Coroutine
        """
        log.debug("Getting client profile")
        device = str(self.device)
        data = {
            "device": device,
            "trusted": True,
            "ver": 90000,
            "includes": "add_by_id,weather,paragraph",
        }
        return self.request(Route("POST", "web/me"), data=data)

    def fetch_friends(
        self, requests: int = 1, dob: bool = True
    ) -> Coroutine[Any, Any, dict[str, Any] | str]:
        """
        Get the list of friends.

        :param requests: The number of friend requests, by default 1.
        :type requests: int, optional
        :param dob: Whether to include date of birth, by default True.
        :type dob: bool, optional
        :return: The coroutine for the request.
        :rtype: Coroutine
        """
        log.debug("Getting friends with requests: %d, dob: %s", requests, dob)
        dob = "true" if dob else "false"
        params = {"requests": requests, "dob": dob, "token": self.token}
        return self.request(Route("GET", "users/me/friends/v2"), params=params)

    def fetch_user_letters(
        self, friend_id: int, page: int = 1
    ) -> Coroutine[Any, Any, dict[str, Any] | str]:
        """
        Fetch letters from a specific friend.

        :param friend_id: The friend's ID.
        :type friend_id: int
        :param page: The page number, by default 1.
        :type page: int, optional
        :return: The coroutine for the request.
        :rtype: Coroutine
        """
        log.debug("Fetching letters for friend_id: %d, page: %d", friend_id, page)
        params = {"token": self.token, "page": page}
        return self.request(Route("GET", f"friend/{friend_id}/all"), params=params)

    async def fetch_auth_passcode(
        self, email: str
    ) -> Coroutine[Any, Any, dict[str, Any] | str]:
        """
        Fetch the authentication passcode.

        :param email: The email address.
        :type email: str
        :return: The coroutine for the request.
        :rtype: Coroutine
        """
        log.debug("Fetching passcode for email: %s", email)
        data = {"email": email, "device": self.device, "checkpass": False}
        return self.request(Route("POST", "auth/email/passcode"), data=data)

    async def fetch_auth_token(
        self, email: str, passcode: str
    ) -> Coroutine[Any, Any, dict[str, Any] | str]:
        """
        Fetch the authentication token.

        :param email: The email address.
        :type email: str
        :param passcode: The passcode.
        :type passcode: str
        :return: The coroutine for the request.
        :rtype: Coroutine
        """
        log.debug("Fetching token for email: %s with passcode: %s", email, passcode)
        data = {"email": email, "passcode": passcode, "device": self.device}
        return self.request(Route("POST", "auth/email"), data=data)



import abc


class Snowflake(metaclass=abc.ABCMeta):

    __slots__ = ()

    @property
    @abc.abstractmethod
    def created_at(self):
        """:class:`datetime.datetime`: Returns the model's creation time as a naive datetime in UTC."""
        raise NotImplementedError

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Snowflake:
            mro = C.__mro__
            for attr in ("created_at", "id"):
                for base in mro:
                    if attr in base.__dict__:
                        break
                else:
                    return NotImplemented
            return True
        return NotImplemented


class User(metaclass=abc.ABCMeta): ...


class Letter(metaclass=abc.ABCMeta): ...



from datetime import datetime
import logging

from . import abc
from ..state import ConnectionState

log = logging.getLogger(__name__)


class BaseLetter(abc.Letter):
    __slots__ = (
        "_state",
        "attachments",
        "avatar",
        "body",
        "created_at",
        "deliver_at",
        "fav",
        "gender",
        "id",
        "location_code",
        "name",
        "post",
        "read_at",
        "sent_from",
        "stamp",
        "status",
        "style",
        "type",
        "updated_at",
        "user",
        "user_fav",
        "user_to",
        "user_to_fav",
    )

    def __init__(self, state: ConnectionState, *, data):
        self._state: ConnectionState = state
        Letter._update(self, data)

    def __str__(self):
        return "<Letter from={0.name!r}>".format(self)

    def _update(self, data):
        for attr in BaseLetter.__slots__:
            if attr.startswith("_"):
                continue
            if attr in [
                "created_at",
                "updated_at",
                "delivered_at",
                "read_at",
            ] and data.get(attr):
                date_object = datetime.strptime(data[attr], "%Y-%m-%d %H:%M:%S")
                setattr(self, attr, date_object)
                continue
            setattr(self, attr, data.get(attr, None))

    def __repr__(self):
        return "<Letter id={0.id!r}>".format(self)


class Letter(BaseLetter):
    def __init__(self, state: ConnectionState, *, data):
        super().__init__(state, data=data)


class AsyncLetterIterator:
    def __init__(self, state, user_id):
        self.state = state
        self.user_id = user_id
        self.current_page = 1
        self.letter_batch = []
        self.next_page = None

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.letter_batch:
            return self.letter_batch.pop(0)

        if self.current_page == 1 or self.next_page:
            data = await self.state.http.fetch_user_letters(
                self.user_id, page=self.current_page
            )
            self.current_page += 1
            self.next_page = data["comments"]["next_page_url"]
            self.letter_batch = [
                Letter(self.state, data=letter) for letter in data["comments"]["data"]
            ]
        if self.letter_batch:
            return self.letter_batch.pop(0)
        else:
            raise StopAsyncIteration

    async def flatten(self) -> list[Letter]:
        all_letters = []
        async for letter in self:
            all_letters.append(letter)
        return all_letters
